# (c) 2023 Sormannilab and Aubin Ramon
#
# Util functions to model and superimpose structures, used in AbNatiV upon humanisation
#
# ============================================================================

from ..model.alignment.aho_consensus import cdr1_aho_indices, cdr2_aho_indices, cdr3_aho_indices, fr_aho_indices
from ..model.alignment.mybio import renumber_Fv_pdb_file

import numpy as np
import os
import Bio



def rmsd(native_atoms: np.array, model_atoms: np.array) -> float:
    """Compute the RMSD between two np arrays of Bio.PDB.Atom atoms."""
    dist = native_atoms - model_atoms
    RMSD = np.sqrt(np.sum(dist**2) / dist.shape[0])
    return RMSD



def merge_list(listo: list) -> np.array:
    """Merge all the lists of a list together into a numpy array"""
    l = []
    [l.extend(i) for i in listo]
    return np.array(l)




def get_atoms(
    model: Bio.PDB.Model,
    atoms_to_be_aligned: list,
    chain_id: str = "H",
    with_side_chain: bool = False,
    backbone_atom_types: list = ["CA", "C", "N"],
) -> dict:
    """Get all BIO Python.atoms of atoms that have an id in the list of ids of interest,
    useful for Bio.PDB.Superimposer() or to compute RMSDs. The output might be shorter
    than the list of id of interests then. But it keeps track of who is who in the keys of
    the output dictionary.

    Parameters
    ----------
        - model: Bio.PDB.Model,
        - atoms_to_be_aligned: list of int
            List of residues numbers to keep the atoms of. Would need to
            be AHo numbers for AHo numbered PDB files.
        - chain_id: str
        - with_side_chain: bool
            If True will consider every heavy atom of a residue.
        - backbone_atom_types: list of str
            List of atoms types to take keep. For backbone should be ['CA','C','N']

    Returns
    -------
        - a dict as follows: {(H,14): [atom_CA, atom_N]}"""

    ca_atoms = dict()
    chain = model[chain_id]

    for res in chain:
        # Check if residue number (.get_id()) is in the list
        if res.get_id()[1] in atoms_to_be_aligned:
            if with_side_chain:
                ca_atoms[(res.resname, res.get_id()[1])] = [
                    atom for atom in res if "H" not in atom.name
                ]
            else:
                ca_atoms[(res.resname, res.get_id()[1])] = [
                    res[atom_type] for atom_type in backbone_atom_types
                ]

    return ca_atoms




def compute_rmsd_of_interest(
    model1: Bio.PDB.Model,
    res_nbs_1: list,
    chain1: str,
    model2: Bio.PDB.Model,
    res_nbs_2: list,
    chain2: str,
    with_side_chain: bool = False,
    backbone_atom_types: list = ["CA", "C", "N"],
) -> float:
    
    """Compute the RMSD between two sets of residues of interest from two different models.
    Options to select side chain or how to define the backbone are available.

    Parameters
    ----------
        - model1: Bio.PDB.Model,
        - res_nbs_1: list of int
            List of residues numbers to keep the atoms of in model1. Would need to
            be AHo numbers for AHo numbered PDB files.
        - chain1: str
        - model2: Bio.PDB.Model,
        - res_nbs_2: list of int
            List of residues numbers to keep the atoms of in model2. Would need to
            be AHo numbers for AHo numbered PDB files.
        - chain2: str
        - with_side_chain: bool
            If True will consider every heavy atom of a residue.
        - backbone_atom_types: list of str
            List of atoms types to consider. Relevant only when with_side_chain=False).
             e.g., for backbone should be ['CA','C','N'].

    Returns
    -------
        - rmsd as a float"""

    # Remove residues that are not the same when compute on the side chain
    if with_side_chain:
        residues_in_1 = [res.get_id()[1] for res in model1[chain1]]
        residues_in_2 = [res.get_id()[1] for res in model2[chain2]]
        new_res_nbs_1, new_res_nbs_2 = list(), list()
        for res1, res2 in zip(res_nbs_1,res_nbs_2):
            if res1 in residues_in_1 and res2 in residues_in_2:
                if model1[chain1][res1].resname == model2[chain2][res2].resname:
                    new_res_nbs_1.append(res1)
                    new_res_nbs_2.append(res2)
        res_nbs_1, res_nbs_2 = new_res_nbs_1, new_res_nbs_2

    model1_atoms = merge_list(
        get_atoms(
            model1, res_nbs_1, chain1, with_side_chain, backbone_atom_types
        ).values()
    )
    model2_atoms = merge_list(
        get_atoms(
            model2, res_nbs_2, chain2, with_side_chain, backbone_atom_types
        ).values()
    )

    if len(model1_atoms) != len(model2_atoms):
        raise ValueError(
            f"Cannot compute RMSD. For the residue numbering given, Model1 has {len(model1_atoms)} selected atoms while Model2 has {len(model2_atoms)} selected atoms."
        )

    return round(rmsd(model1_atoms, model2_atoms), 4)



def superimpose_two_models(
    fixed_model: Bio.PDB.Model,
    chain_vh_fixed: str,
    chain_vl_fixed: str,
    res_nbs_vh_fixed: list,
    res_nbs_vl_fixed: list,
    moving_model: Bio.PDB.Model,
    chain_vh_moving: str,
    chain_vl_moving: str,
    res_nbs_vh_moving: list,
    res_nbs_vl_moving: list,
    is_VHH: bool,
    take_union_ng: bool,
    fp_save_aligned_model: str,
    with_side_chain: bool = False,
    backbone_atom_types: list = ["CA", "C", "N"],
) -> float:
    """
    Superimpose a moving model on a fixed model on a set of chosen residues (via their residue number).

    If models are AHo numbered, there is take_union_ng option to take the union sets of residues with the same AHo number.
    Useful when aligning on the framework region and want to take the union of non gaps residues. If models are not
    AHo numebred this will not work.

    Save the aligned version of the moving model in fp_save_aligned_model.

    Parameters
    ----------
        - fixed_model: Bio.PDB.Model,
        - chain_vh_fixed: str,
        - chain_vl_fixed: str,
        - res_nbs_vh_fixed: list of int
            List of residues numbers to keep the atoms of in fixed model heavy chain. Would need to
            be AHo numbers for AHo numbered PDB files.
        - res_nbs_vl_fixed: list of int
            List of residues numbers to keep the atoms of in fixed model light chain. Would need to
            be AHo numbers for AHo numbered PDB files.
        - moving_model: Bio.PDB.Model,
        - chain_vh_moving: str,
        - chain_vl_moving: str,
        - res_nbs_vh_moving: list of int
            List of residues numbers to keep the atoms of in moving model heavy chain. Would need to
            be AHo numbers for AHo numbered PDB files.
        - res_nbs_vl_moving: list of int
            List of residues numbers to keep the atoms of in moving model light chain. Would need to
            be AHo numbers for AHo numbered PDB files.
        - is_VHH: bool
            Will consider the light chain if False.
        - take_union_ng: bool
            If True, take the union of residues with the same res number. Useful when aligning numbered models,
            like for framework for isntance. If models are not AHo numebred this will not work.
        - fp_save_aligned_model: str
            File path where to save the aligned moving model.
        - with_side_chain: bool
            If True will consider every heavy atom of a residue when selecting the atoms.
        - backbone_atom_types: list of str
            List of atoms types to consider. Relevant only when with_side_chain=False).
             e.g., for backbone should be ['CA','C','N'].


    Returns
    -------
        - RMSD upon alignemnt on the set of given AHo positions.
    """

    # Get atom scaffolds
    fixed_vh_ca_atoms = get_atoms(
        fixed_model,
        res_nbs_vh_fixed,
        chain_vh_fixed,
        with_side_chain,
        backbone_atom_types,
    )
    moving_vh_ca_atoms = get_atoms(
        moving_model,
        res_nbs_vh_moving,
        chain_vh_moving,
        with_side_chain,
        backbone_atom_types,
    )

    if not is_VHH:
        fixed_vl_ca_atoms = get_atoms(
            fixed_model,
            res_nbs_vl_fixed,
            chain_vl_fixed,
            with_side_chain,
            backbone_atom_types,
        )
        moving_vl_ca_atoms = get_atoms(
            moving_model,
            res_nbs_vl_moving,
            chain_vl_moving,
            with_side_chain,
            backbone_atom_types,
        )

    # Will take only the Union set of no-gap scaffold residues
    if take_union_ng:
        union_fixed_atoms, union_moving_atoms = list(), list()

        for fixed_res_name, fixed_res_id in fixed_vh_ca_atoms:
            if (fixed_res_name, fixed_res_id) in moving_vh_ca_atoms:
                union_fixed_atoms.extend(
                    fixed_vh_ca_atoms[(fixed_res_name, fixed_res_id)]
                )
                union_moving_atoms.extend(
                    moving_vh_ca_atoms[(fixed_res_name, fixed_res_id)]
                )

        if not is_VHH:
            for fixed_res_name, fixed_res_id in fixed_vl_ca_atoms:
                if (fixed_res_name, fixed_res_id) in moving_vl_ca_atoms:
                    union_fixed_atoms.extend(
                        fixed_vl_ca_atoms[(fixed_res_name, fixed_res_id)]
                    )
                    union_moving_atoms.extend(
                        moving_vl_ca_atoms[(fixed_res_name, fixed_res_id)]
                    )

    else:
        union_fixed_atoms, union_moving_atoms = merge_list(
            fixed_vh_ca_atoms.values()
        ), merge_list(moving_vh_ca_atoms.values())

        if not is_VHH:
            union_fixed_atoms = np.append(
                union_fixed_atoms, merge_list(fixed_vl_ca_atoms.values())
            )
            union_moving_atoms = np.append(
                union_moving_atoms, merge_list(moving_vl_ca_atoms.values())
            )

    # Superimposition on set of
    super_imposer = Bio.PDB.Superimposer()
    super_imposer.set_atoms(union_fixed_atoms, union_moving_atoms)
    super_imposer.apply(moving_model.get_atoms())

    align_rmsd = round(super_imposer.rms, 4)

    io = Bio.PDB.PDBIO()
    io.set_structure(moving_model)
    io.save(fp_save_aligned_model)

    return round(align_rmsd, 4)




def compute_cdr_displacement_between_two_structures(name_seq1:str, fp_pdb1: str, h_chain1: str, l_chain1: str, do_renumb1: bool,
                                            name_seq2:str, fp_pdb2:str, h_chain2: str, l_chain2: str, do_renumb2: bool,
                                            is_vhh: bool=True, is_abnativ2: bool=True,
                                            o_dir_pdb_models:str='./')->float:
    
    '''Superimpose two structures (models or crystals) on their scaffold and compute the CDR displacement (RSMD).

    Parameters:
    -----------
    name_seq1 : str
        Name or identifier of the first sequence/structure (used for labeling in output).
    fp_pdb1 : str
        File path to the first PDB structure.
    h_chain1 : str
        Heavy-chain identifier in the first structure (e.g. 'H' or 'A').
    l_chain1 : str
        Light-chain identifier in the first structure (ignored if `is_vhh=True`).
    do_renumb1 : bool
        Whether to renumber the first structure to the AHo numbering scheme.

    name_seq2 : str
        Name or identifier of the second sequence/structure (used for labeling in output).
    fp_pdb2 : str
        File path to the second PDB structure.
    h_chain2 : str
        Heavy-chain identifier in the second structure.
    l_chain2 : str
        Light-chain identifier in the second structure (ignored if `is_vhh=True`).
    do_renumb2 : bool
        Whether to renumber the second structure to the AHo numbering scheme.

    is_vhh : bool, optional
        Whether the input structures are nanobodies (single-domain VHHs). 
        If True, light-chain parameters are ignored. Default is True.
    is_abnativ2 : bool, optional
        If True, check for AHo CDR gaps during renumbering (used for compatibility with AbNatiV2 pipelines).
        Default is True.
    o_dir_pdb_models : str, optional
        Output directory where the aligned model will be saved as a PDB file.
        Default is the current directory ('./').

    Returns
    -------
    float or tuple of floats
        - For VHHs (`is_vhh=True`): Returns RMSD values for Scaffold, H-CDR1, H-CDR2, and H-CDR3.  
        - For Fvs (`is_vhh=False`): Returns RMSD values for Scaffold, H-CDR1–3 and L-CDR1–3.
    '''
    
    # Renumber in AHo numbering if needed, useful when one is a crystal structure
    if do_renumb1:
        aho_fp_pdb1 = os.path.join(os.path.dirname(fp_pdb1), os.path.basename(fp_pdb1).split('.')[0] + '_aho.pdb')
        renumber_Fv_pdb_file(fp_pdb1, h_chain1, l_chain1, is_VHH=is_vhh, scheme='AHo', outfilename=aho_fp_pdb1, check_AHo_CDR_gaps=is_abnativ2)
    else: aho_fp_pdb1 = fp_pdb1
    
    if do_renumb2:
        aho_fp_pdb2 = os.path.join(os.path.dirname(fp_pdb2), os.path.basename(fp_pdb2).split('.')[0] + '_aho.pdb')
        renumber_Fv_pdb_file(fp_pdb2, h_chain2, l_chain2, is_VHH=is_vhh, scheme='AHo', outfilename=aho_fp_pdb2, check_AHo_CDR_gaps=is_abnativ2)
    else: aho_fp_pdb2 = fp_pdb2

    model1 = Bio.PDB.PDBParser().get_structure(name_seq1, aho_fp_pdb1)[0]
    model2 = Bio.PDB.PDBParser().get_structure(name_seq2, aho_fp_pdb2)[0]

    # Superimpose the frameworks
    fp_save_aligned_model = os.path.join(o_dir_pdb_models,f'{name_seq2}_aligned_with_{name_seq1}.pdb')
    scaffold_rmsd = superimpose_two_models(model1, h_chain1, l_chain1, fr_aho_indices, fr_aho_indices,
                            model2, h_chain2, l_chain2, fr_aho_indices, fr_aho_indices,
                            is_VHH=is_vhh, take_union_ng=True,fp_save_aligned_model=fp_save_aligned_model)
    print(f'-> RMSD on scaffold upon alignment on scaffold:\n--> {scaffold_rmsd}A')

    # Compute RMSD on CDRs
    h_cdr1 = round(compute_rmsd_of_interest(model1, cdr1_aho_indices, h_chain1, model2, cdr1_aho_indices, h_chain2), 4)
    h_cdr2 = round(compute_rmsd_of_interest(model1, cdr2_aho_indices, h_chain1, model2, cdr2_aho_indices, h_chain2), 4)
    h_cdr3 = round(compute_rmsd_of_interest(model1, cdr3_aho_indices, h_chain1, model2, cdr3_aho_indices, h_chain2), 4)

    if not is_vhh:
        l_cdr1 = round(compute_rmsd_of_interest(model1, cdr1_aho_indices, l_chain1, model2, cdr1_aho_indices, l_chain2), 4)
        l_cdr2 = round(compute_rmsd_of_interest(model1, cdr2_aho_indices, l_chain1, model2, cdr2_aho_indices, l_chain2), 4)
        l_cdr3 = round(compute_rmsd_of_interest(model1, cdr3_aho_indices, l_chain1, model2, cdr3_aho_indices, l_chain2), 4)

        print(f'-> RMSD on CDRs upon alignment on scaffold:\n--> H-CDR1: {h_cdr1:.2f}A / H-CDR2: {h_cdr2:.2f}A / H-CDR3: {h_cdr3:.2f}A\n--> L-CDR1: {l_cdr1:.2f}A / L-CDR2: {l_cdr2:.2f}A / L-CDR3: {l_cdr3:.2f}A\n')
        return fp_save_aligned_model, scaffold_rmsd, h_cdr1, h_cdr2, h_cdr3, l_cdr1, l_cdr2, l_cdr3

    else:
        print(f'-> RMSD on CDRs upon alignment on scaffold:\n--> H-CDR1: {h_cdr1:.2f}A / H-CDR2: {h_cdr2:.2f}A / H-CDR3: {h_cdr3:.2f}A\n')
        return fp_save_aligned_model, scaffold_rmsd, h_cdr1, h_cdr2, h_cdr3
    


# (c) 2023 Sormannilab and Aubin Ramon
#
# Scoring functions for the AbNatiV model.
#
# ============================================================================
import os

# os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"   # allow CPU fallback for missing MPS ops

from .abnativ2_unpaired import AbNatiV_Model as AbNatiV2_Model
from .abnativ2_paired import AbNatiV_Paired_Model
from .abnativ import AbNatiV_Model

from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from importlib import resources
from .onehotencoder import (
    data_loader_masking_bert_onehot,
    data_loader_masking_bert_onehot_paired,
    alphabet,
)
from .alignment.mybio import anarci_alignments_of_Fv_sequences_iter
from .alignment.aho_consensus import (
    cdr1_aho_indices,
    cdr2_aho_indices,
    cdr3_aho_indices,
    fr_aho_indices,
)
from ..init import PRETRAINED_MODELS_DIR

from typing import Tuple
from collections import defaultdict
from tqdm import tqdm
import math
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import torch

from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq


def plot_abnativ_profile(res_scores, full_seq, name_seq, model_type, fig_fp_save):
    """
    Plot the AbNatiV score profile of a sequence.
    """
    sns.set(font_scale=2.2)
    sns.set_style(
        "white",
        {
            "axes.spines.right": False,
            "axes.spines.top": True,
            "axes.spines.bottom": False,
            "xtick.bottom": False,
            "xtick.top": True,
            "ytick.left": True,
            "xtick.labeltop": True,
        },
    )
    fig, ax = plt.subplots(figsize=(40, 8))

    # Plot
    ax.plot(res_scores, linewidth=5, alpha=0.65, color="darkorange", label=name_seq)

    # Add CDRs
    ax.axvspan(
        cdr1_aho_indices[0] - 1,
        cdr1_aho_indices[-1] - 1,
        alpha=0.06,
        color="forestgreen",
    )
    ax.axvspan(
        cdr2_aho_indices[0] - 1,
        cdr2_aho_indices[-1] - 1,
        alpha=0.06,
        color="forestgreen",
    )
    ax.axvspan(
        cdr3_aho_indices[0] - 1,
        cdr3_aho_indices[-1] - 1,
        alpha=0.06,
        color="forestgreen",
    )

    ax.xaxis.set_ticks(np.arange(0, len(full_seq), 1.0))
    ax.set_xticklabels(full_seq, fontsize=21)
    ax.tick_params(axis="x", which="major", pad=3)
    ax.xaxis.set_label_position("top")
    ax.set_ylabel(f"AbNatiV {model_type}\nResidue Score", fontsize=28, labelpad=15)
    ax.set_xlabel("Sequence", fontsize=28, labelpad=15)
    ax.xaxis.tick_top()

    plt.title(f"{name_seq} AbNatiV {model_type} profile", fontsize=31, pad=18)
    plt.tight_layout()
    plt.savefig(fig_fp_save, dpi=200, bbox_inches="tight")


def norm_by_length_no_gap(scores_pposi: list, onehot_encoding: list) -> list:
    """
    Sum the scores per position of a given sequence. Normalise the result by the
    number of residues (no gaps).

    Parameters
    ----------
        - scores_pposi : list
            Scores of the chosen positions
        - onehot_encoding : list
            One-hot encoding of the chosen given positions
    """
    length = len(onehot_encoding)
    idx = np.argmax(onehot_encoding, axis=1)
    nb_gaps = np.count_nonzero(idx == 20)
    length -= nb_gaps
    if length == 0:
        print("Length portion equals zero, too many gaps. Evaluates it as np.nan")
        norm = np.nan
    else:
        norm = np.sum(scores_pposi) / length
    return norm


def linear_rescaling(list_scores: list, t_new: int, t_r: int) -> list:
    """
    Linear rescaling of the scores to translate the threshold t_r (specific to each
    model type as defined in Methods) to the new threshold t_new.
    """
    rescaled = list()
    for x in list_scores:
        rescaled.append((t_new - 1) / (t_r - 1) * (x - 1) + 1)
    return rescaled

def percentile_of_score(df_scores: pd.DataFrame, score_label: str, new_scores: np.array):
    """
    Compute percentile scores"""
    distribution = np.sort(df_scores[score_label])
    return (np.searchsorted(distribution, new_scores, side='right') / len(distribution) * 100).astype(int)

def get_abnativ_nativeness_scores(
    output_abnativ: dict, portion_indices: list, model_type: str
) -> list:
    """
    Give the AbNatiV nativeness scores for given positions (could be full) of the outputs of the
    AbNatiV model.

    Parameters
    ----------
        - output_abnativ : dict
            Output dict of sequences evaluated by AbNatiV
            i.e. {'fp':'/my/path/to/the/dataset.txt', 'recon_error_pbe': [0.2,0.3,0.4]}
        - portion_indices : list
            Position indices to score
            e.g., range(1,150) for the all sequence / range(27,43) for the CDR-1 (AHo numbering)
        - model_type : str
            VH, VHH, VKappa, VLambda

    Returns
    ----------
        - a list of the rescaled AbNatiV nativenees scores
    """

    humanness_scores = list()
    best_thresholds = {
        "VH": 0.988047,
        "VKappa": 0.992496,
        "VLambda": 0.985580,
        "VHH": 0.990973,
        "VH2": 0.993624,
        "VL2": 0.995629,
        "VHH2": 0.992695,
        "VPaired2": 0.991817,
        "VPairedH2": 0.992019,
        "VPairedL2": 0.993152,
        "VH2-Rhesus": 0.993892,
    }

    if "2" in model_type:
        output_data_to_compute_one = "mse_pposi"
    else:
        output_data_to_compute_one = "recon_error_pposi"

    # Convert AHo-position into list index
    portion_indices = np.array(portion_indices) - 1

    # Move tensors to NumPy before looping
    output_profiles = output_abnativ[output_data_to_compute_one].detach().cpu().numpy()
    input_profiles = output_abnativ["inputs"].detach().cpu().numpy()

    # Extract relevant portions in batch
    output_profiles_portion = output_profiles[:, portion_indices]
    input_profiles_portion = input_profiles[:, portion_indices]

    # Compute scores in a vectorized manner
    scores = np.array(
        [
            norm_by_length_no_gap(out, inp)
            for out, inp in zip(output_profiles_portion, input_profiles_portion)
        ]
    )

    # Apply exponential transformation
    humanness_scores.extend(np.exp(-scores))

    if model_type not in best_thresholds.keys():  # If scoring your own model
        rescaled_scores = humanness_scores

    else:
        rescaled_scores = linear_rescaling(
            humanness_scores, 0.8, best_thresholds[model_type]
        )

    return list(rescaled_scores)


@torch.inference_mode()
def abnativ_scoring(
    model_type: str,
    seq_records: list,
    batch_size: int = 1024,
    mean_score_only: bool = True,
    do_align: bool = True,
    is_VHH: bool = False,
    is_plotting_profiles: bool = False,
    output_dir: str = './',
    output_id: str = "antibody",
    verbose: bool = True,
    run_parall_al: int = False,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Infer on a list of seuqences, the AbNatiV loaded model with the
    selected model type. Returns a dataframe with the scored sequences.
    Alignement (AHo numbering) is performed if asked. If not asked, the sequences must have been
    beforehand aligned on the AHo scheme.

    Parameters
    ----------
        - model_type: str
            e.g., - VH, VHH, VKappa, VLambda (for default AbNatiV trained models),
                  - or, filepath to the custom checkpoint .ckpt (no linear rescaling will be applied)
        - seq_records: list
            List of SeqRecords from the BioPython package. seq = str(record.seq) / id = record.id
        - batch_size: int
        - mean_score_only: bool
            If True, provide only the mean nativeness score at the sequence level
        - do_align: bool
            If True, do the alignment with the AHo numbering by ANARCI #Coming update. A column
            will be added with the aligned_seq
        - is_VHH: bool
            If True, considers the VHH seed for the alignment, more suitable when aligning nanobody sequences
        - is_plotting_profiles: bool
            If True, plot the profile of every sequence
        - output_dir: str
            Filepath to the folder whre all files are saved
        - id: str
            Preffix of all saved files
        - verbose: bool
            If False, do not print anything except exceptions and errors
        - run_parall_al: int or bool
            If int, will parralelise the alignment on that int of cpus.
            If you don't want to parallelise, enter False

    Returns
    -------
        - df_mean: pd.DataFrame
            Dataframe composed of the id, the aligned sequence, the AbNatiV overall score,
            and in particular the CDR-1, CDR-2, CDR-3, Framework scores for a each sequences of the fasta file
        - df_profile: pd_DataFrame
            If not mean_score_only: Dataframe composed of the residue level Abnativ score with the score of each residue at each position.
            Else: Empty Dataframe

    """
    # Set the device
    if torch.cuda.is_available():
        device_type = "cuda"
    elif hasattr(torch.backends, "mps") and torch.backends.mps.is_available():
        device_type = 'mps'
    else:
        device_type = "cpu"

    if verbose:
        print(f"\nCalculations on device {device_type}\n")

    device = torch.device(device_type)

    model_dir = PRETRAINED_MODELS_DIR

    fr_trained_models = {  # AbNatiV1 models
        "VH": os.path.join(model_dir, "vh_model.ckpt"),
        "VHH": os.path.join(model_dir, "vhh_model.ckpt"),
        "VKappa": os.path.join(model_dir, "vkappa_model.ckpt"),
        "VLambda": os.path.join(model_dir, "vlambda_model.ckpt"),
        # AbNatiV2 models
        "VH2": os.path.join(model_dir, "vh2_model.ckpt"),
        "VL2": os.path.join(model_dir, "vl2_model.ckpt"),
        "VHH2": os.path.join(model_dir, "vhh2_model.ckpt"),
        # AbNatiV2-Zoo models
        "VH2-Rhesus": os.path.join(model_dir, "vh2_rhesus_model.ckpt"),
    }

    # Create folder if not existing
    if not os.path.exists(output_dir) and is_plotting_profiles:
        os.makedirs(output_dir)

    ## ALIGNMENT ##
    if do_align:
        if "2" in model_type:
            check_AHo_CDR_gaps = True
        else:
            check_AHo_CDR_gaps = False

        VH, VK, VL, failed, mischtype = anarci_alignments_of_Fv_sequences_iter(
            seq_records,
            isVHH=is_VHH,
            verbose=verbose,
            run_parallel=run_parall_al,
            check_AHo_CDR_gaps=check_AHo_CDR_gaps,
            del_cyst_misalign=False,
            check_terms=False,
        )

        recs = VH.to_recs()
        recs.extend(VK.to_recs())
        recs.extend(VL.to_recs())

        list_ids = [rec.id for rec in recs]
        list_al_seqs = [str(rec.seq) for rec in recs]
    else:
        list_al_seqs = [str(rec.seq) for rec in seq_records]
        list_ids = [rec.id for rec in seq_records]
        for k, al_seq in enumerate(list_al_seqs):
            if len(al_seq) != 149:
                raise Exception(
                    f"Sequence {k} is not aligned as expected (length={len(al_seq)}!=149), make sure all sequences are aligned on the AHo numbering."
                )

    list_seqs = [seq.replace("-", "") for seq in list_al_seqs]

    ## MODEL LOADING ##
    if model_type not in fr_trained_models.keys():
        try:
            if verbose:
                print(
                    f"\n### AbNatiV scoring of {output_id} from checkpoint {model_type} ###\n"
                )
            loaded_model = AbNatiV_Model.load_from_checkpoint(
                model_type, map_location=device
            )
            name_type = "Custom"
        except:
            print(
                f"Cannnot load the checkpoint {model_type}, you might use the default models: VH, VKappa, VLambda, or VHH."
            )

    else:
        if verbose:
            print(f"\n### AbNatiV {model_type}-ness scoring of {output_id} ###\n")
        if "2" in model_type:
            loaded_model = AbNatiV2_Model.load_from_checkpoint(
                fr_trained_models[model_type], map_location=device
            )
        else:
            loaded_model = AbNatiV_Model.load_from_checkpoint(
                fr_trained_models[model_type], map_location=device
            )
        name_type = model_type

    loaded_model.eval()
    loader = data_loader_masking_bert_onehot(
        list_al_seqs, batch_size, perc_masked_residues=0, is_masking=False
    )

    nb_of_iterations = math.ceil(len(list_ids) / batch_size)

    scored_data_dict_mean = defaultdict(list)
    scored_data_dict_mean.update(
        {"seq_id": list_ids, "input_seq": list_seqs, "aligned_seq": list_al_seqs}
    )

    scored_data_dict_profile = defaultdict(list)

    # Percentiles for V2
    if '2' in name_type:
        try:
            from importlib import resources
            percentile_dir = resources.files(__package__).joinpath("percentiles")
        except AttributeError:
            import importlib_resources
            percentile_dir = importlib_resources.files(__package__).joinpath("percentiles")

        fp_test_scores = {
            "VH2": os.path.join(percentile_dir, "test_vh_v2_abnativ_seq_scores.csv"),
            "VL2": os.path.join(percentile_dir, "test_vl_v2_abnativ_seq_scores.csv"),
            "VHH2": os.path.join(percentile_dir, "test_vhh_v2_abnativ_seq_scores.csv"),
        }

        df_test_scores = pd.read_csv(fp_test_scores[name_type])

    ## MODEL EVALUATION ##
    seen = 0
    for count, batch in enumerate(
        tqdm(loader, total=nb_of_iterations, disable=not verbose)
    ):
        batch = move_to_device(batch, device)
        output_abnativ = loaded_model(batch)

        # Sequence-level scores
        humanness_scores = get_abnativ_nativeness_scores(
            output_abnativ, range(1, 150), model_type
        )
        cdr1_scores = get_abnativ_nativeness_scores(
            output_abnativ, cdr1_aho_indices, model_type
        )
        cdr2_scores = get_abnativ_nativeness_scores(
            output_abnativ, cdr2_aho_indices, model_type
        )
        cdr3_scores = get_abnativ_nativeness_scores(
            output_abnativ, cdr3_aho_indices, model_type
        )
        fr_scores = get_abnativ_nativeness_scores(
            output_abnativ, fr_aho_indices, model_type
        )

        scored_data_dict_mean[f"AbNatiV {name_type} Score"].extend(humanness_scores)

        scored_data_dict_mean[f"AbNatiV CDR1-{name_type} Score"].extend(cdr1_scores)
        scored_data_dict_mean[f"AbNatiV CDR2-{name_type} Score"].extend(cdr2_scores)
        scored_data_dict_mean[f"AbNatiV CDR3-{name_type} Score"].extend(cdr3_scores)
        scored_data_dict_mean[f"AbNatiV FR-{name_type} Score"].extend(fr_scores)

        # Percentiles for V2 to test set
        if '2' in name_type:
            scored_data_dict_mean[f"AbNatiV {name_type} Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV {name_type} Score", humanness_scores))
            scored_data_dict_mean[f"AbNatiV CDR1-{name_type} Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV CDR1-{name_type} Score", cdr1_scores))
            scored_data_dict_mean[f"AbNatiV CDR2-{name_type} Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV CDR2-{name_type} Score", cdr2_scores))
            scored_data_dict_mean[f"AbNatiV CDR3-{name_type} Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV CDR3-{name_type} Score", cdr3_scores))
            scored_data_dict_mean[f"AbNatiV FR-{name_type} Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV FR-{name_type} Score", fr_scores))

        # Residue-level scores
        if not mean_score_only:
            batch_inputs = (
                output_abnativ["inputs"].cpu().numpy()
            )  # Move tensor to NumPy array
            batch_recon = output_abnativ["x_recon"].detach().cpu().numpy()

            err_key = "mse_pposi" if "2" in name_type else "recon_error_pposi"
            batch_errors = np.exp(
                -output_abnativ[err_key].detach().cpu().numpy()
            )  # Compute exp(-error)

            batch_size, seq_len, _ = (
                batch_recon.shape
            )  # Get batch size and sequence length

            seq_ids = list_ids[seen : seen + batch_size]  # Slice corresponding seq_ids
            seen += batch_size

            # Get max indices for amino acids
            aa_indices = np.argmax(batch_inputs, axis=2)  # Shape: (batch_size, seq_len)
            aa_chars = np.array(alphabet)[aa_indices]  # Convert indices to characters

            # Revert to 'X' for clarity in csv and png profiles
            mask_X = np.isclose(
                batch_inputs, 1.0 / len(alphabet), atol=1e-6, rtol=0
            ).all(axis=2)
            aa_chars[mask_X] = "X"

            # Flatten arrays for efficient DataFrame creation
            seq_id_expanded = np.repeat(seq_ids, seq_len)
            position_expanded = np.tile(np.arange(1, seq_len + 1), batch_size)
            aa_expanded = aa_chars.flatten()
            scores_expanded = batch_errors.flatten()

            scored_data_dict_profile["seq_id"].extend(seq_id_expanded)
            scored_data_dict_profile["AHo position"].extend(position_expanded)
            scored_data_dict_profile["aa"].extend(aa_expanded)
            scored_data_dict_profile[f"AbNatiV {name_type} Residue Score"].extend(
                scores_expanded
            )

            # Store reconstruction probabilities for each amino acid (one-hot encoded)
            for i, aa in enumerate(alphabet):
                scored_data_dict_profile[aa].extend(batch_recon[:, :, i].flatten())

            if is_plotting_profiles:
                dir_save = os.path.join(output_dir, f"{output_id}_profiles")
                os.makedirs(dir_save, exist_ok=True)  # Efficient directory creation

                for k, seq_id in enumerate(seq_ids):
                    clean_seq_id = seq_id.replace("/", "_")
                    fig_save_fp = os.path.join(
                        dir_save, f"{clean_seq_id}_abnativ_profile.png"
                    )
                    plot_abnativ_profile(
                        batch_errors[k], aa_chars[k], seq_id, name_type, fig_save_fp
                    )

    df_mean = pd.DataFrame.from_dict(scored_data_dict_mean)
    df_profile = pd.DataFrame.from_dict(scored_data_dict_profile)

    return df_mean, df_profile


def move_to_device(data, device):
    if isinstance(data, torch.Tensor):
        return data.to(device)
    elif isinstance(data, list):
        return [move_to_device(item, device) for item in data]
    elif isinstance(data, tuple):
        return tuple(move_to_device(item, device) for item in data)
    else:
        raise TypeError(f"Unsupported data type: {type(data)}")


def abnativ_scoring_csv(
    model_type: str,
    input_csv: str,
    col_id: str = "name",
    col_seq: str = "sequence",
    batch_size: int = 1024,
    chunk_size: int = 150000,
    mean_score_only: bool = True,
    do_align: bool = True,
    is_VHH: bool = False,
    is_plotting_profiles: bool = False,
    output_dir: str = './',
    output_id: str = "antibody",
    verbose: bool = True,
    run_parall_al: int = False,
    preserve_columns: bool = True,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Score sequences from a CSV file, preserving all original columns.
    Supports chunked processing for large files to manage memory efficiently.

    Parameters
    ----------
        - model_type: str
            e.g., - VH, VHH, VKappa, VLambda (for default AbNatiV trained models),
                  - VH2, VL2, VHH2 (for AbNatiV2 models),
                  - or, filepath to the custom checkpoint .ckpt (no linear rescaling will be applied)
        - input_csv: str
            Filepath to the input CSV file
        - col_id: str
            Name of the column containing sequence identifiers (default: "name")
        - col_seq: str
            Name of the column containing sequences (default: "sequence")
        - batch_size: int
            Batch size for model inference (default: 1024)
        - chunk_size: int
            Number of sequences to process per chunk for memory management (default: 150000).
            Set to 0 or None to disable chunking.
        - mean_score_only: bool
            If True, provide only the mean nativeness score at the sequence level
        - do_align: bool
            If True, do the alignment with the AHo numbering by ANARCI
        - is_VHH: bool
            If True, considers the VHH seed for the alignment, more suitable when aligning nanobody sequences
        - is_plotting_profiles: bool
            If True, plot the profile of every sequence
        - output_dir: str
            Filepath to the folder where all files are saved
        - output_id: str
            Prefix of all saved files
        - verbose: bool
            If False, do not print anything except exceptions and errors
        - run_parall_al: int or bool
            If int, will parallelise the alignment on that int of cpus.
            If you don't want to parallelise, enter False
        - preserve_columns: bool
            If True, preserve all original columns from the input CSV in the output (default: True)

    Returns
    -------
        - df_mean: pd.DataFrame
            Dataframe with AbNatiV scores merged with original CSV columns (if preserve_columns=True)
        - df_profile: pd.DataFrame
            If not mean_score_only: Dataframe with residue-level scores.
            Else: Empty Dataframe

    """
    import gc
    import os

    # Read input CSV
    if verbose:
        print(f"\n### Loading CSV file: {input_csv} ###\n")

    df_input = pd.read_csv(input_csv)

    # Validate columns exist
    if col_id not in df_input.columns:
        raise ValueError(f"ID column '{col_id}' not found in CSV. Available columns: {list(df_input.columns)}")
    if col_seq not in df_input.columns:
        raise ValueError(f"Sequence column '{col_seq}' not found in CSV. Available columns: {list(df_input.columns)}")

    # Drop rows with missing values in required columns
    df_input = df_input.dropna(subset=[col_id, col_seq])

    if verbose:
        print(f"Loaded {len(df_input)} sequences from CSV")

    # Build SeqRecords
    seq_records = [
        SeqRecord(Seq(str(seq)), id=str(name), name=str(name), description="")
        for seq, name in zip(df_input[col_seq], df_input[col_id])
    ]

    total_seqs = len(seq_records)

    # Determine if chunking is needed
    if chunk_size and chunk_size > 0 and total_seqs > chunk_size:
        num_chunks = (total_seqs + chunk_size - 1) // chunk_size
        if verbose:
            print(f"Processing {total_seqs} sequences in {num_chunks} chunk(s) of up to {chunk_size}")

        all_mean_dfs = []
        all_profile_dfs = []

        for chunk_idx in range(num_chunks):
            start_idx = chunk_idx * chunk_size
            end_idx = min((chunk_idx + 1) * chunk_size, total_seqs)

            chunk_records = seq_records[start_idx:end_idx]
            if verbose:
                print(f"\n  Processing chunk {chunk_idx + 1}/{num_chunks}: sequences {start_idx} to {end_idx - 1} ({len(chunk_records)} seqs)")

            # Score this chunk
            chunk_mean_df, chunk_profile_df = abnativ_scoring(
                model_type=model_type,
                seq_records=chunk_records,
                batch_size=batch_size,
                mean_score_only=mean_score_only,
                do_align=do_align,
                is_VHH=is_VHH,
                is_plotting_profiles=is_plotting_profiles,
                output_dir=output_dir,
                output_id=f"{output_id}_chunk{chunk_idx}",
                verbose=verbose,
                run_parall_al=run_parall_al,
            )

            all_mean_dfs.append(chunk_mean_df)
            if not mean_score_only:
                all_profile_dfs.append(chunk_profile_df)

            # Free memory
            del chunk_mean_df, chunk_profile_df
            gc.collect()

        # Combine all chunks
        if verbose:
            print(f"\nCombining {len(all_mean_dfs)} chunk results...")

        df_mean = pd.concat(all_mean_dfs, ignore_index=True)
        df_profile = pd.concat(all_profile_dfs, ignore_index=True) if all_profile_dfs else pd.DataFrame()

        del all_mean_dfs, all_profile_dfs
        gc.collect()

    else:
        # Process all at once
        df_mean, df_profile = abnativ_scoring(
            model_type=model_type,
            seq_records=seq_records,
            batch_size=batch_size,
            mean_score_only=mean_score_only,
            do_align=do_align,
            is_VHH=is_VHH,
            is_plotting_profiles=is_plotting_profiles,
            output_dir=output_dir,
            output_id=output_id,
            verbose=verbose,
            run_parall_al=run_parall_al,
        )

    # Merge with original CSV to preserve all columns
    if preserve_columns:
        if verbose:
            print(f"\nMerging scores with original CSV columns...")

        # Create a mapping from seq_id to original row index for proper merging
        # The df_mean has 'seq_id' column from abnativ_scoring
        df_input_indexed = df_input.copy()
        df_input_indexed['_merge_key'] = df_input_indexed[col_id].astype(str)

        # The scores df has 'seq_id' which corresponds to col_id values
        df_mean['_merge_key'] = df_mean['seq_id'].astype(str)

        # Get score columns (everything except seq_id, input_seq, aligned_seq, _merge_key)
        score_columns = [col for col in df_mean.columns if col not in ['seq_id', 'input_seq', '_merge_key']]

        # Merge: keep all original columns + add score columns
        df_merged = df_input_indexed.merge(
            df_mean[['_merge_key'] + score_columns],
            on='_merge_key',
            how='left'
        )

        # Drop the merge key
        df_merged = df_merged.drop(columns=['_merge_key'])

        df_mean = df_merged

    return df_mean, df_profile


@torch.inference_mode()
def abnativ_scoring_paired(
    df_pairs: pd.DataFrame,
    col_id: str = "ID",
    col_vh: str = "vh_seq",
    col_vl: str = "vl_seq",
    batch_size: int = 256,
    mean_score_only: bool = True,
    do_align: bool = True,
    is_plotting_profiles: bool = False,
    output_dir: str = "./",
    output_id: str = "antibody",
    verbose: bool = True,
    run_parall_al: int = False,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Infer on a list of seuqences, the AbNatiV loaded model with the
    selected model type. Returns a dataframe with the scored sequences.
    Alignement (AHo numbering) is performed if asked. If not asked, the sequences must have been
    beforehand aligned on the AHo scheme.

    Parameters
    ----------
        - df_pairs: pd.DataFrame
            Dataframe with paired sequences
        - col_id: str
            Name of the column with the sequence IDs
        - col_vh: str
            Name of the column with the heavy sequences
        - col_vl: str:
            Name of the column with the light sequences
        - batch_size: int
        - mean_score_only: bool
            If True, provide only the mean nativeness score at the sequence level
        - do_align: bool
            If True, do the alignment with the AHo numbering by ANARCI #Coming update. A column
            will be added with the aligned_seq
        - is_plotting_profiles: bool
            If True, plot the profile of every sequence
        - output_dir: str
            Filepath to the folder whre all files are saved
        - id: str
            Preffix of all saved files
        - verbose: bool
            If False, do not print anything except exceptions and errors
        - run_parall_al: int or bool
            If int, will parralelise the alignment on that int of cpus.
            If you don't want to parallelise, enter False

    Returns
    -------
        - df_mean: pd.DataFrame
            Dataframe composed of the id, the aligned sequence, the AbNatiV overall score,
            and in particular the CDR-1, CDR-2, CDR-3, Framework scores for a each sequences of the fasta file
        - df_profile: pd_DataFrame
            If not mean_score_only: Dataframe composed of the residue level Abnativ score with the score of each residue at each position.
            Else: Empty Dataframe
            For paired sequences the returning profile concatenates the heavy and the light chains together.

    """
    # Set the device
    if torch.cuda.is_available():
        device_type = "cuda"
    elif hasattr(torch.backends, "mps") and torch.backends.mps.is_available():
        device_type = 'mps'
    else:
        device_type = "cpu"

    if verbose:
        print(f"\nCalculations on device {device_type}\n")

    device = torch.device(device_type)

    model_dir = PRETRAINED_MODELS_DIR

    # AbNatiV2 paired model
    fr_trained_model = os.path.join(model_dir, "vpaired2_model.ckpt")

    # Create folder if not existing
    if not os.path.exists(output_dir) and is_plotting_profiles:
        os.makedirs(output_dir)

    ## ALIGNMENT ENSURING DISCARDED PAIRED SEQUENCES WHEN ONLY ONE SEQUENCE PASSES THE ALIGNMENT ##
    if do_align:
        # Heavy
        seq_records_h = [
            SeqRecord(Seq(df_pairs[col_vh].iloc[i]), id=df_pairs[col_id].iloc[i])
            for i in range(len(df_pairs))
        ]
        VH, VK, VL, failed, mischtype = anarci_alignments_of_Fv_sequences_iter(
            seq_records_h,
            isVHH=False,
            verbose=verbose,
            run_parallel=run_parall_al,
            check_AHo_CDR_gaps=True,
            del_cyst_misalign=False,
            check_terms=False,
        )
        recs_h = VH.to_recs()
        recs_h.extend(VK.to_recs())
        recs_h.extend(VL.to_recs())

        # Light
        seq_records_l = [
            SeqRecord(Seq(df_pairs[col_vl].iloc[i]), id=df_pairs[col_id].iloc[i])
            for i in range(len(df_pairs))
        ]
        VH, VK, VL, failed, mischtype = anarci_alignments_of_Fv_sequences_iter(
            seq_records_l,
            isVHH=False,
            verbose=verbose,
            run_parallel=run_parall_al,
            check_AHo_CDR_gaps=True,
            del_cyst_misalign=False,
            check_terms=False,
        )
        recs_l = VH.to_recs()
        recs_l.extend(VK.to_recs())
        recs_l.extend(VL.to_recs())

        # Pair the alignment
        sequences = {"H": {}, "L": {}}
        for record in recs_h:
            sequences["H"][record.id] = str(record.seq)
        for record in recs_l:
            sequences["L"][record.id] = str(record.seq)

        paired_vh, paired_vl, ids = list(), list(), list()
        for seq_id in sequences["H"].keys():
            if seq_id in sequences["L"]:
                ids.append(seq_id)
                paired_vh.append(sequences["H"][seq_id])
                paired_vl.append(sequences["L"][seq_id])

        col_id_al, col_vh_al, col_vl_al = "ID", "vh_seq_aligned", "vl_seq_aligned"
        d = {col_id_al: ids, col_vh_al: paired_vh, col_vl_al: paired_vl}
        df_pairs_al = pd.DataFrame(d)

    else:
        df_pairs_al = df_pairs
        col_id_al, col_vh_al, col_vl_al = col_id, col_vh, col_vl
        for k, al_seq in enumerate(df_pairs_al[col_vh_al]):
            if len(al_seq) != 149:
                raise Exception(
                    f"Sequence {k} is not aligned as expected (length={len(al_seq)}!=149), make sure all sequences are aligned on the AHo numbering."
                )
        for k, al_seq in enumerate(df_pairs_al[col_vl_al]):
            if len(al_seq) != 149:
                raise Exception(
                    f"Sequence {k} is not aligned as expected (length={len(al_seq)}!=149), make sure all sequences are aligned on the AHo numbering."
                )

    ## MODEL LOADING ##
    loaded_model = AbNatiV_Paired_Model.load_from_checkpoint(
        fr_trained_model, map_location=device
    )
    loaded_model.eval()

    loader = data_loader_masking_bert_onehot_paired(
        df_pairs_al,
        col_vh_al,
        col_vl_al,
        batch_size,
        perc_masked_residues=0,
        is_masking=False,
    )

    nb_of_iterations = math.ceil(len(df_pairs_al) / batch_size)

    scored_data_dict_mean = defaultdict(list)
    scored_data_dict_mean.update(
        {
            "seq_id": df_pairs_al[col_id_al],
            "input_seq_vh": [seq.replace("-", "") for seq in df_pairs_al[col_vh_al]],
            "input_seq_vl": [seq.replace("-", "") for seq in df_pairs_al[col_vl_al]],
            "aligned_seq_vh": df_pairs_al[col_vh_al],
            "aligned_seq_vl": df_pairs_al[col_vl_al],
            "aligned_seq": [
                h + l for h, l in zip(df_pairs_al[col_vh_al], df_pairs_al[col_vl_al])
            ],
        }
    )

    scored_data_dict_profile = defaultdict(list)

    # Percentiles for V2
    try:
        from importlib import resources
        percentile_dir = resources.files(__package__).joinpath("percentiles")
    except AttributeError:
        import importlib_resources
        percentile_dir = importlib_resources.files(__package__).joinpath("percentiles")
    df_test_scores = pd.read_csv(os.path.join(percentile_dir, "test_vpaired_v2_abnativ_seq_scores.csv"))

    ## MODEL EVALUATION ##
    seen = 0
    for count, batch in enumerate(
        tqdm(loader, total=nb_of_iterations, disable=not verbose)
    ):
        batch = move_to_device(batch, device)
        output_abnativ = loaded_model(batch)

        # Sequence-level scores
        whole_humanness_scores = get_abnativ_nativeness_scores(
            output_abnativ, range(1, 299), "VPaired2"
        )
        pairing_pred_scores = (
            output_abnativ["pairing_pred"].cpu().detach().numpy().flatten()
        )

        h_humanness_scores = get_abnativ_nativeness_scores(
            output_abnativ, range(1, 150), "VPairedH2"
        )
        l_humanness_scores = get_abnativ_nativeness_scores(
            output_abnativ, range(150, 299), "VPairedL2"
        )

        h_cdr1_scores = get_abnativ_nativeness_scores(
            output_abnativ, cdr1_aho_indices, "VPairedH2"
        )
        h_cdr2_scores = get_abnativ_nativeness_scores(
            output_abnativ, cdr2_aho_indices, "VPairedH2"
        )
        h_cdr3_scores = get_abnativ_nativeness_scores(
            output_abnativ, cdr3_aho_indices, "VPairedH2"
        )
        h_fr_scores = get_abnativ_nativeness_scores(
            output_abnativ, fr_aho_indices, "VPairedH2"
        )

        l_cdr1_scores = get_abnativ_nativeness_scores(
            output_abnativ, np.array(cdr1_aho_indices) + 149, "VPairedL2"
        )
        l_cdr2_scores = get_abnativ_nativeness_scores(
            output_abnativ, np.array(cdr2_aho_indices) + 149, "VPairedL2"
        )
        l_cdr3_scores = get_abnativ_nativeness_scores(
            output_abnativ, np.array(cdr3_aho_indices) + 149, "VPairedL2"
        )
        l_fr_scores = get_abnativ_nativeness_scores(
            output_abnativ, np.array(fr_aho_indices) + 149, "VPairedL2"
        )

        scored_data_dict_mean[f"AbNatiV Heavy-Light Score"].extend(
            whole_humanness_scores
        )
        scored_data_dict_mean[f"AbNatiV Pairing Score (%)"].extend(pairing_pred_scores)

        scored_data_dict_mean[f"AbNatiV Heavy Score"].extend(h_humanness_scores)
        scored_data_dict_mean[f"AbNatiV Light Score"].extend(l_humanness_scores)

        scored_data_dict_mean[f"AbNatiV CDR1-Heavy Score"].extend(h_cdr1_scores)
        scored_data_dict_mean[f"AbNatiV CDR2-Heavy Score"].extend(h_cdr2_scores)
        scored_data_dict_mean[f"AbNatiV CDR3-Heavy Score"].extend(h_cdr3_scores)
        scored_data_dict_mean[f"AbNatiV FR-Heavy Score"].extend(h_fr_scores)

        scored_data_dict_mean[f"AbNatiV CDR1-Light Score"].extend(l_cdr1_scores)
        scored_data_dict_mean[f"AbNatiV CDR2-Light Score"].extend(l_cdr2_scores)
        scored_data_dict_mean[f"AbNatiV CDR3-Light Score"].extend(l_cdr3_scores)
        scored_data_dict_mean[f"AbNatiV FR-Light Score"].extend(l_fr_scores)

        # Percentiles for V2 to test set
        scored_data_dict_mean[f"AbNatiV Heavy-Light Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV Heavy-Light Score", whole_humanness_scores))

        scored_data_dict_mean[f"AbNatiV CDR1-Heavy Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV CDR1-Heavy Score", h_cdr1_scores))
        scored_data_dict_mean[f"AbNatiV CDR2-Heavy Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV CDR2-Heavy Score", h_cdr2_scores))
        scored_data_dict_mean[f"AbNatiV CDR3-Heavy Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV CDR3-Heavy Score", h_cdr3_scores))
        scored_data_dict_mean[f"AbNatiV FR-Heavy Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV FR-Heavy Score", h_fr_scores))
        
        scored_data_dict_mean[f"AbNatiV CDR1-Light Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV CDR1-Light Score", l_cdr1_scores))
        scored_data_dict_mean[f"AbNatiV CDR2-Light Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV CDR2-Light Score", l_cdr2_scores))
        scored_data_dict_mean[f"AbNatiV CDR3-Light Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV CDR3-Light Score", l_cdr3_scores))
        scored_data_dict_mean[f"AbNatiV FR-Light Percentile"].extend(percentile_of_score(df_test_scores, f"AbNatiV FR-Light Score", l_fr_scores))


        # Residue-level scores
        if not mean_score_only:
            # Move tensors to NumPy in one step
            batch_inputs = output_abnativ["inputs"].cpu().numpy()
            batch_recon = output_abnativ["x_recon"].detach().cpu().numpy()
            batch_errors = np.exp(
                -output_abnativ["mse_pposi"].detach().cpu().numpy()
            )  # Precompute exp(-error)

            batch_size, seq_len, _ = batch_recon.shape  # Get dimensions

            seq_ids = (
                df_pairs_al[col_id_al].iloc[seen : seen + batch_size].to_numpy()
            )  # Slice corresponding seq_ids
            seen += batch_size

            # Get max indices for amino acids (faster than looping)
            aa_indices = np.argmax(batch_inputs, axis=2)
            aa_chars = np.array(alphabet)[aa_indices]

            # Revert to 'X' for clarity in csv and png profiles
            mask_X = np.isclose(
                batch_inputs, 1.0 / len(alphabet), atol=1e-6, rtol=0
            ).all(axis=2)
            aa_chars[mask_X] = "X"

            # Flatten arrays for efficient DataFrame creation
            seq_id_expanded = np.repeat(seq_ids, seq_len)
            position_expanded = np.tile(np.arange(1, seq_len + 1), batch_size)
            aa_expanded = aa_chars.flatten()
            scores_expanded = batch_errors.flatten()

            aho_labels = np.where(
                position_expanded > 149,
                [f"L-{pos - 149}" for pos in position_expanded],
                [f"H-{pos}" for pos in position_expanded],
            )

            scored_data_dict_profile["seq_id"].extend(seq_id_expanded)
            scored_data_dict_profile["AHo position"].extend(aho_labels)
            scored_data_dict_profile["aa"].extend(aa_expanded)
            scored_data_dict_profile[f"AbNatiV VPaired2 Residue Score"].extend(
                scores_expanded
            )

            # Store reconstruction probabilities for each amino acid
            for i, aa in enumerate(alphabet):
                scored_data_dict_profile[aa].extend(batch_recon[:, :, i].flatten())

            if is_plotting_profiles:
                dir_save = os.path.join(output_dir, f"{output_id}_profiles")
                os.makedirs(dir_save, exist_ok=True)  # Efficient directory creation

                for k, seq_id in enumerate(seq_ids):
                    clean_seq_id = seq_id.replace("/", "_")
                    fig_save_fp = os.path.join(
                        dir_save, f"{clean_seq_id}_abnativ_paired_heavy_profile.png"
                    )
                    plot_abnativ_profile(
                        batch_errors[k, :149],
                        aa_chars[k, :149],
                        seq_id,
                        "H-VPaired2",
                        fig_save_fp,
                    )

                    fig_save_fp = os.path.join(
                        dir_save, f"{clean_seq_id}_abnativ_paired_light_profile.png"
                    )
                    plot_abnativ_profile(
                        batch_errors[k, 149:],
                        aa_chars[k, 149:],
                        seq_id,
                        "L-VPaired2",
                        fig_save_fp,
                    )

    df_mean = pd.DataFrame.from_dict(scored_data_dict_mean)
    df_profile = pd.DataFrame.from_dict(scored_data_dict_profile)

    return df_mean, df_profile



# (c) 2023 Sormannilab and Aubin Ramon
#
# Lightning testing of the AbNatiV model.
#
# ============================================================================

from .model.scoring_functions import abnativ_scoring_paired
from .model.onehotencoder import alphabet
import argparse
import os
import pandas as pd

from .model.utils import is_protein
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq

from .init import PRETRAINED_MODELS_DIR

def run(args: argparse.Namespace):

    # Check that models are downloaded
    if not os.path.exists(PRETRAINED_MODELS_DIR):
        raise Exception("Models not found. Please run 'abnativ update' to download the models.")

    if not os.path.exists(args.output_directory):
        os.makedirs(args.output_directory)

    ## DATA SCORING ##
    batch_size = 128

    # Load the sequences
    if is_protein(args.vh, alphabet) and is_protein(args.vl, alphabet):
        df_pairs = pd.DataFrame({'ID': ['single_seq'], 'vh_seq': [args.vh], 'vl_seq': [args.vl]})
        col_id, col_vh, col_vl = 'ID', 'vh_seq', 'vl_seq'

    elif os.path.isfile(args.input_filepath_csv):
        df_pairs = pd.read_csv(args.input_filepath_csv)
        col_id, col_vh, col_vl = args.col_id, args.col_vh, args.col_vl

    else: raise Exception(f'Can not find the input file {args.input_filepath_or_seq} or if you gave a single vh and vl protein sequences make sure it is composed only of {alphabet}')

    abnativ_df_mean, abnativ_df_profile = abnativ_scoring_paired(df_pairs, col_id, col_vh, col_vl, batch_size, args.mean_score_only,
                                                          args.do_align, args.is_plotting_profiles, args.output_directory, args.output_id,
                                                          run_parall_al=args.ncpu)

    ## DATA SAVING ##
    print(f'\n-> Scores being saved in {args.output_directory}\n')
    abnativ_df_mean.to_csv(os.path.join(args.output_directory, f'{args.output_id}_abnativ_seq_scores.csv'))
    if not args.mean_score_only:
        abnativ_df_profile.to_csv(os.path.join(args.output_directory, f'{args.output_id}_abnativ_res_scores.csv'))
    if args.is_plotting_profiles:
        save_profile_fp = os.path.join(args.output_directory, f'{args.output_id}_profiles')
        print(f'\n-> Profile plots saved in {save_profile_fp}\n')









# (c) 2023 Sormannilab and Aubin Ramon
#
# Diverse functions to run AbNatiV.
#
# ============================================================================

from typing import Tuple
import math

import torch
from torch import nn, einsum
import torch.nn.functional as F
import torch.distributed as distributed

from einops import rearrange, repeat

def is_protein(seq, aa_list):
    """
    Check if a str corresponds to a protein sequence
    return bool
    """
    for aa in seq:
        if aa not in aa_list:
            return False
    return True

def l_out_cnn1d(L_in:int,K:int,S:int,P:int,D:int=1) -> float:
    '''Formula to find the L_out dimension of an input (dim=L_in)
    in cnn_1d.'''
    return (L_in+2*P-D*(K-1)-1)/S + 1

def find_optimal_cnn1d_padding(L_in:int,K,S:int) -> Tuple[int,int]:
    '''Find the minimal padding giving the kernel size K and stride S 
    for a CNN1D without losing any piece of information.'''
    P=0
    L_out = l_out_cnn1d(L_in,K,S,P)

    assert L_in>=K, 'Kernel size higher than input dimension, the conv1d will not work'

    while not L_out.is_integer() and 2*P<=S:
        L_out = l_out_cnn1d(L_in,K,S,P)
        P+=1

    if 2*P>=S: P-=1
    return math.floor(L_out), P

def l_out_cnn1d_transpose(L_in:int,K:int,S:int,P:int,D:int=1) -> int:
    '''Formula to find the L_out dimension of an input (dim=L_in)
    in cnn_1d.'''
    return (L_in-1)*S -2*P + D*(K-1) + 1

def find_out_padding_cnn1d_transpose(L_obj:int,L_in:int,K:int,S:int,P:int) -> int:
    '''Find the minimal output padding giving the kernel size K and stride S 
    to add after a CNN1D transpose layer to reach L_obj (objective).'''
    L_out = l_out_cnn1d_transpose(L_in,K,S,P)
    assert L_obj>=L_out, 'Make sure the padding is correct, the ouput \
            of the CNN1D transpose is larger than expeceted'
    return L_obj-L_out

# From the enhancing VQ (https://github.com/lucidrains/vector-quantize-pytorch/blob/master/vector_quantize_pytorch/vector_quantize_pytorch.py)
# Copyright (c) 2020 Phil Wang (MIT Licenced)

def exists(val):
    return val is not None

def default(val, d):
    return val if exists(val) else d

def noop(*args, **kwargs):
    pass

def l2norm(t):
    return F.normalize(t, p = 2, dim = -1)

def log(t, eps = 1e-20):
    return torch.log(t.clamp(min = eps))

def uniform_init(*shape):
    t = torch.empty(shape)
    nn.init.kaiming_uniform_(t)
    return t

def gumbel_noise(t):
    noise = torch.zeros_like(t).uniform_(0, 1)
    return -log(-log(noise))

def gumbel_sample(t, temperature = 1., dim = -1):
    if temperature == 0:
        return t.argmax(dim = dim)

    return ((t / temperature) + gumbel_noise(t)).argmax(dim = dim)

def ema_inplace(moving_avg, new, decay):
    moving_avg.data.mul_(decay).add_(new, alpha = (1 - decay))

def laplace_smoothing(x, n_categories, eps = 1e-5):
    return (x + eps) / (x.sum() + n_categories * eps)

def sample_vectors(samples, num):
    num_samples, device = samples.shape[0], samples.device
    if num_samples >= num:
        indices = torch.randperm(num_samples, device = device)[:num]
    else:
        indices = torch.randint(0, num_samples, (num,), device = device)

    return samples[indices]

def batched_sample_vectors(samples, num):
    return torch.stack([sample_vectors(sample, num) for sample in samples.unbind(dim = 0)], dim = 0)

def pad_shape(shape, size, dim = 0):
    return [size if i == dim else s for i, s in enumerate(shape)]

def sample_multinomial(total_count, probs):
    device = probs.device
    probs = probs.cpu()

    total_count = probs.new_full((), total_count)
    remainder = probs.new_ones(())
    sample = torch.empty_like(probs, dtype = torch.long)

    for i, p in enumerate(probs):
        s = torch.binomial(total_count, p / remainder)
        sample[i] = s
        total_count -= s
        remainder -= p

    return sample.to(device)

def all_gather_sizes(x, dim):
    size = torch.tensor(x.shape[dim], dtype = torch.long, device = x.device)
    all_sizes = [torch.empty_like(size) for _ in range(distributed.get_world_size())]
    distributed.all_gather(all_sizes, size)
    return torch.stack(all_sizes)

def all_gather_variably_sized(x, sizes, dim = 0):
    rank = distributed.get_rank()
    all_x = []

    for i, size in enumerate(sizes):
        t = x if i == rank else x.new_empty(pad_shape(x.shape, size, dim))
        distributed.broadcast(t, src = i, async_op = True)
        all_x.append(t)

    distributed.barrier()
    return all_x

def sample_vectors_distributed(local_samples, num):
    local_samples = rearrange(local_samples, '1 ... -> ...')

    rank = distributed.get_rank()
    all_num_samples = all_gather_sizes(local_samples, dim = 0)

    if rank == 0:
        samples_per_rank = sample_multinomial(num, all_num_samples / all_num_samples.sum())
    else:
        samples_per_rank = torch.empty_like(all_num_samples)

    distributed.broadcast(samples_per_rank, src = 0)
    samples_per_rank = samples_per_rank.tolist()

    local_samples = sample_vectors(local_samples, samples_per_rank[rank])
    all_samples = all_gather_variably_sized(local_samples, samples_per_rank, dim = 0)
    out = torch.cat(all_samples, dim = 0)

    return rearrange(out, '... -> 1 ...')

def batched_bincount(x, *, minlength):
    batch, dtype, device = x.shape[0], x.dtype, x.device
    target = torch.zeros(batch, minlength, dtype = dtype, device = device)
    values = torch.ones_like(x)
    target.scatter_add_(-1, x, values)
    return target

def kmeans(
    samples,
    num_clusters,
    num_iters = 10,
    use_cosine_sim = False,
    sample_fn = batched_sample_vectors,
    all_reduce_fn = noop
):
    num_codebooks, dim, dtype, device = samples.shape[0], samples.shape[-1], samples.dtype, samples.device

    means = sample_fn(samples, num_clusters)

    for _ in range(num_iters):
        if use_cosine_sim:
            dists = samples @ rearrange(means, 'h n d -> h d n')
        else:
            dists = -torch.cdist(samples, means, p = 2)

        buckets = torch.argmax(dists, dim = -1)
        bins = batched_bincount(buckets, minlength = num_clusters)
        all_reduce_fn(bins)

        zero_mask = bins == 0
        bins_min_clamped = bins.masked_fill(zero_mask, 1)

        new_means = buckets.new_zeros(num_codebooks, num_clusters, dim, dtype = dtype)

        new_means.scatter_add_(1, repeat(buckets, 'h n -> h n d', d = dim), samples)
        new_means = new_means / rearrange(bins_min_clamped, '... -> ... 1')
        all_reduce_fn(new_means)

        if use_cosine_sim:
            new_means = l2norm(new_means)

        means = torch.where(
            rearrange(zero_mask, '... -> ... 1'),
            means,
            new_means
        )

    return means, bins

def batched_embedding(indices, embeds):
    batch, dim = indices.shape[1], embeds.shape[-1]
    indices = repeat(indices, 'h b n -> h b n d', d = dim)
    embeds = repeat(embeds, 'h c d -> h b c d', b = batch)
    return embeds.gather(2, indices)

# regularization losses

def orthogonal_loss_fn(t):
    # eq (2) from https://arxiv.org/abs/2112.00384
    h, n = t.shape[:2]
    normed_codes = l2norm(t)
    cosine_sim = einsum('h i d, h j d -> h i j', normed_codes, normed_codes)
    return (cosine_sim ** 2).sum() / (h * n ** 2) - (1 / n)



"""
 Copyright 2023. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0
"""

import argparse
import sys
from . import init
from . import scoring
from . import paired_scoring
from . import training
from . import vhh_humanisation
from . import vh_vl_humanisation
from . import vh_vl_paired_humanisation

USAGE = """

%(prog)s <command> [options]

AbNatiV provides three commands:
    - init: download the default AbNatiV models to the latest version or specifies a tag to checkout
    - train: train AbNatiV on a new input dataset of antibody sequences
    - score: use a trained AbNatiV model (default or custom) to score a set of input antibody sequences
            (supports FASTA, single sequence, or CSV input with configurable column names)
    - paired_score: use the paired AbNatiV2 model to score a set of input antibody sequences
    - hum_vhh: humanise a VHH sequence by combining AbNatiV VH and VHH assessments (dual-control strategy)
    - hum_vh_vl: humanise a pair of VH/VL Fv sequences by increasing AbNatiV VH- and VL- humanness
    - hum_vh_vl_paired: humanise a pair of VH/VL Fv sequences by increasing AbNatiV VH- and VL- humanness jointly with the paired model pAbNatiV2

see also
%(prog)s <command> score -h
%(prog)s <command> paired_score -h
%(prog)s <command> train -h
%(prog)s <command> hum_vhh -h
%(prog)s <command> hum_vh_vl -h
%(prog)s <command> hum_vh_vl_paired -h
for additional help

Copyright 2023. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0
"""


def main():

    if len(sys.argv) == 1:
        empty_parser = argparse.ArgumentParser(
            description="VQ-VAE-based assessment of antibody and nanobody nativeness for engineering, selection, and computational design",
            usage=USAGE
        )
        empty_parser.print_help(sys.stderr)
        sys.exit(1)

    parser = argparse.ArgumentParser(
        description="VQ-VAE-based assessment of antibody and nanobody nativeness for engineering, selection, and computational design",
    )

    subparser = parser.add_subparsers()

    # INIT
    update_parser = subparser.add_parser("init", description="Initialise the default AbNatiV models to the latest version or specifies a tag to checkout",
                                        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    update_parser.add_argument('-fu', '--force_update', help='Force redownloaded all models', action="store_true")

    # update_parser.add_argument('-fp', '--fp_save', help='Filepath where to save the model weights. If "" (default), it will save them in your root home folder in .abnativ', type=str, default="")

    update_parser.set_defaults(func=lambda args: init.init(args))


    # TRAINING
    train_parser = subparser.add_parser("train",
                                        description="Train AbNatiV on a new input dataset of antibody sequences",
                                        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    train_parser.add_argument('-tr', '--train_filepath', help='Filepath to fasta file .fa with sequences for training', type=str,
                        default='train_2M.fa')

    train_parser.add_argument('-va', '--val_filepath', help='Filepath to fasta file .fa with sequences for validation', type=str,
                        default='val_50k.fa')

    train_parser.add_argument('-hp', '--hparams', help='Filepath to the hyperparameter dictionary .yml', type=str,
                        default='hparams.yml')

    train_parser.add_argument('-mn', '--model_name', help='Name of the model to save checkpoints in', type=str,
                        default='abnativ_v2')

    train_parser.add_argument('-align', '--do_align', help='Do the alignment and the cleaning of the given sequences before training. \
                              This step can takes a lot of time if the number of sequences is huge.', action="store_true")

    train_parser.add_argument('-ncpu', '--ncpu', help='If ncpu>1 will parallelise the algnment process', default=1)

    train_parser.add_argument('-isVHH', '--is_VHH', help='Considers the VHH seed for the alignment/ \
                              It is more suitable when aligning nanobody sequences', action="store_true")

    train_parser.set_defaults(func=lambda args: training.run(args))


    # SCORING
    score_parser = subparser.add_parser("score", description="Use a trained AbNatiV model (default or custom) to score a set of input antibody sequences",
                                        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    score_parser.add_argument('-nat', '--nativeness_type', help='To load the AbNatiV default trained models type VH, VKappa, VLambda, or VHH, \
                              otherwise add directly the path to your own AbNatiV trained checkpoint .ckpt', type=str,
                                default='VH')

    score_parser.add_argument('-mean', '--mean_score_only', help='Generate only a file with a score per sequence. \
                              If not, generate a second file with a nativeness score per positin with a probability score \
                              for each aa at each position.', action="store_true")

    score_parser.add_argument('-i', '--input_filepath_or_seq', help='Filepath to the fasta file .fa to score or directly \
                              a single string sequence', type=str,
                              default='to_score.fa')

    score_parser.add_argument('-odir', '--output_directory', help='Filepath of the folder where all files are saved', type=str,
                              default='abnativ_scoring')

    score_parser.add_argument('-oid', '--output_id', help='Prefix of all the saved filenames (e.g., name sequence)', type=str,
                              default='antibody_vh')

    score_parser.add_argument('-align', '--do_align', help='Do the alignment and the cleaning of the given sequences before scoring. \
                              This step can takes a lot of time if the number of sequences is huge.', action="store_true")

    score_parser.add_argument('-ncpu', '--ncpu', help='If ncpu>1 will parallelise the algnment process', type=int, default=1)

    score_parser.add_argument('-isVHH', '--is_VHH', help='Considers the VHH seed for the alignment. \
                              It is more suitable when aligning nanobody sequences', action="store_true")

    score_parser.add_argument('-plot', '--is_plotting_profiles', help='Plot profile for every input sequence and save them in {output_directory}/{output_id}_profiles.', action="store_true")

    # CSV-specific arguments
    score_parser.add_argument('-cid', '--col_id', help='Name of the ID/name column in the CSV file (only used for CSV input)', type=str,
                              default='name')

    score_parser.add_argument('-cseq', '--col_seq', help='Name of the sequence column in the CSV file (only used for CSV input)', type=str,
                              default='sequence')

    score_parser.add_argument('-chunk', '--chunk_size', help='Number of sequences to process per chunk for memory management with large CSV files. \
                              Set to 0 to disable chunking. (only used for CSV input)', type=int,
                              default=150000)

    score_parser.set_defaults(func=lambda args: scoring.run(args))


    # PAIRED_SCORING
    score_parser = subparser.add_parser("paired_score", description="Use the paired AbNatiV2 model to score a set of input antibody sequences",
                                        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    score_parser.add_argument('-i', '--input_filepath_csv', help='Filepath to the csv file .csv to score sequences. \
                              Leave empty if running directly a sequence without a csv file.', type=str,
                              default='to_score.fa')

    score_parser.add_argument('-vh', '--vh', help='VH sequence to score directly. \
                              Leave empty if running directly a csv file.', type=str,
                              default='no_seq.')

    score_parser.add_argument('-vl', '--vl', help='VL sequence to score directly. \
                              Leave empty if running directly a csv file.', type=str,
                              default='no_seq.')

    score_parser.add_argument('-cid', '--col_id', help='Name of the ID column of the csv file. \
                              Leave empty if running directly a sequence without a csv file.', type=str,
                              default='ID')

    score_parser.add_argument('-cvh', '--col_vh', help='Name of the vh column of the csv file. \
                              Leave empty if running directly a sequence without a csv file.', type=str,
                              default='vh_seq')

    score_parser.add_argument('-cvl', '--col_vl', help='Name of the vl column of the csv file. \
                              Leave empty if running directly a sequence without a csv file.', type=str,
                              default='vl_seq')

    score_parser.add_argument('-mean', '--mean_score_only', help='Generate only a file with a score per sequence. \
                              If not, generate a second file with a nativeness score per positin with a probability score \
                              for each aa at each position.', action="store_true")

    score_parser.add_argument('-odir', '--output_directory', help='Filepath of the folder where all files are saved', type=str,
                              default='abnativ_scoring')

    score_parser.add_argument('-oid', '--output_id', help='Prefix of all the saved filenames (e.g., name sequence)', type=str,
                              default='antibody_vh')

    score_parser.add_argument('-align', '--do_align', help='Do the alignment and the cleaning of the given sequences before scoring. \
                              This step can takes a lot of time if the number of sequences is huge.', action="store_true")

    score_parser.add_argument('-ncpu', '--ncpu', help='If ncpu>1 will parallelise the algnment process', type=int, default=1)

    score_parser.add_argument('-plot', '--is_plotting_profiles', help='Plot profile for every input sequence and save them in {output_directory}/{output_id}_profiles.', action="store_true")

    score_parser.set_defaults(func=lambda args: paired_scoring.run(args))


    # VHH HUMANISATION
    hum_vhh_parser = subparser.add_parser("hum_vhh", description="Use AbNatiV to humanise nanobody sequences by combining AbNatiV VH and VHH assessments (dual-control stategy).",
                                        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    hum_vhh_parser.add_argument('-i', '--input_filepath_or_seq', help='Filepath to the fasta file .fa to score or directly \
                              a single string sequence', type=str,
                              default='to_score.fa')

    hum_vhh_parser.add_argument('-vh', '--nat_vh', help='The nativeness score to use for the humanness (VH or VH2)',
                              type=str, default='VH2')

    hum_vhh_parser.add_argument('-vhh', '--nat_vhh', help='The nativeness score to use for the VHH-nativeness (VHH or VHH2)',
                              type=str, default='VHH2')

    hum_vhh_parser.add_argument('-odir', '--output_directory', help='Filepath of the folder where all files are saved', type=str,
                              default='abnativ_humanisation_vhh')

    hum_vhh_parser.add_argument('-oid', '--output_id', help='Prefix of all the saved filenames (e.g., name sequence)', type=str,
                              default='nanobody_vhh')

    hum_vhh_parser.add_argument('-VHscore', '--threshold_abnativ_score', help='Bellow the AbNatiV VH threshold score, a position is considered as a liability',
                              type=float, default=0.98)

    hum_vhh_parser.add_argument('-rasa', '--threshold_rasa_score', help='Above this threshold, the residue is considered solvent exposed and is considered for mutation',
                              type=float, default=0.15)

    hum_vhh_parser.add_argument('-isExhaustive', '--is_Exhaustive', help='If True, runs the Exhaustive sampling strategy. If False, runs the enhanced sampling method', action="store_true")

    hum_vhh_parser.add_argument('-fmut', '--forbidden_mut', help='List of string residues to ban for mutation, i.e. C M', nargs='*',
                              type=str, default=['C','M'])

    hum_vhh_parser.add_argument('-VHHdecrease', '--perc_allowed_decrease_vhh', help='Maximun ΔVHH score decrease allowed for a mutation',
                              type=float, default=5e-2)

    hum_vhh_parser.add_argument('-a', '--a', help='Used for enhanced sampling method in multi-objective selection function: aΔVH+bΔVHH',
                              type=float, default=2)

    hum_vhh_parser.add_argument('-b', '--b', help='Used for enhanced sampling method in multi-objective selection function: aΔVH+bΔVHH',
                              type=float, default=1)

    hum_vhh_parser.add_argument('-pdb', '--pdb_file', help='Filepath to a pdb crystal structure of the nanobody of interest used to compute \
                                the solvent exposure. If the PDB is not very cleaned that might lead to some false results (which should be flagged by the \
                                program). If None, will predict the structure using NanoBuilder2', type=str,
                              default=None)

    hum_vhh_parser.add_argument('-ch', '--ch_id', help='PDB chain id of the nanobody of interest. If -pdb is None, it does not matter', type=str,
                              default='H')

    hum_vhh_parser.set_defaults(func=lambda args: vhh_humanisation.run(args))


    # VH_VL HUMANISATION
    hum_vh_vl_parser = subparser.add_parser("hum_vh_vl", description="Use AbNatiV to humanise a pair of VH/VL Fv sequences by increasing AbNatiV VH- and VL- humanness.",
                                        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    hum_vh_vl_parser.add_argument('-i_vh', '--input_seq_vh', help='A single VH string sequence', type=str)

    hum_vh_vl_parser.add_argument('-i_vl', '--input_seq_vl', help='A single VL string sequence', type=str)

    hum_vh_vl_parser.add_argument('-vh', '--nat_vh', help='The nativeness score to use for the heavy chain (VH or VH2)',
                              type=str, default='VH2')

    hum_vh_vl_parser.add_argument('-vl', '--nat_vl', help='The nativeness score to use for the light chain (VL or VL2)',
                              type=str, default='VL2')

    hum_vh_vl_parser.add_argument('-odir', '--output_directory', help='Filepath of the folder where all files are saved', type=str,
                              default='abnativ_humanisation_vh_vl')

    hum_vh_vl_parser.add_argument('-oid', '--output_id', help='Prefix of all the saved filenames (e.g., name sequence)', type=str,
                              default='antibody_vh_vl')

    hum_vh_vl_parser.add_argument('-score', '--threshold_abnativ_score', help='Bellow the AbNatiV threshold score, a position is considered as a liability',
                              type=float, default=0.98)

    hum_vh_vl_parser.add_argument('-rasa', '--threshold_rasa_score', help='Above this threshold, the residue is considered solvent exposed and is considered for mutation',
                              type=float, default=0.15)

    hum_vh_vl_parser.add_argument('-isExhaustive', '--is_Exhaustive', help='If True, runs the Exhaustive sampling strategy. If False, runs the enhanced sampling method', action="store_true")

    hum_vh_vl_parser.add_argument('-fmut', '--forbidden_mut', help='List of string residues to ban for mutation, i.e. C M', nargs='*',
                              type=str, default=['C','M'])

    hum_vh_vl_parser.add_argument('-pdb', '--pdb_file', help='Filepath to a pdb crystal structure of the nanobody of interest used to compute \
                                the solvent exposure. If the PDB is not very cleaned that might lead to some false results (which should be flagged by the \
                                program). If None, will predict the paired structure using ABodyBuilder2', type=str,
                              default=None)

    hum_vh_vl_parser.add_argument('-ch_vh', '--ch_id_vh', help='PDB chain id of the heavy chain of interest. If -pdb is None, it does not matter', type=str,
                              default='H')

    hum_vh_vl_parser.add_argument('-ch_vl', '--ch_id_vl', help='PDB chain id of the light chain of interest. If -pdb is None, it does not matter', type=str,
                              default='L')

    hum_vh_vl_parser.set_defaults(func=lambda args: vh_vl_humanisation.run(args))



    # VH_VL HUMANISATION PAIRED
    hum_vh_vl_paired_parser = subparser.add_parser("hum_vh_vl_paired", description="Use AbNatiV to humanise a pair of VH/VL Fv sequences by increasing AbNatiV VH- and VL- humanness jointy with the paired model, while improving the pairing likelihood.",
                                        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    hum_vh_vl_paired_parser.add_argument('-i_vh', '--input_seq_vh', help='A single VH string sequence', type=str)

    hum_vh_vl_paired_parser.add_argument('-i_vl', '--input_seq_vl', help='A single VL string sequence', type=str)

    hum_vh_vl_paired_parser.add_argument('-odir', '--output_directory', help='Filepath of the folder where all files are saved', type=str,
                              default='abnativ_humanisation_vh_vl')

    hum_vh_vl_paired_parser.add_argument('-oid', '--output_id', help='Prefix of all the saved filenames (e.g., name sequence)', type=str,
                              default='antibody_vh_vl')

    hum_vh_vl_paired_parser.add_argument('-score', '--threshold_abnativ_score', help='Bellow the AbNatiV threshold score, a position is considered as a liability',
                              type=float, default=0.98)

    hum_vh_vl_paired_parser.add_argument('-rasa', '--threshold_rasa_score', help='Above this threshold, the residue is considered solvent exposed and is considered for mutation',
                              type=float, default=0.15)

    hum_vh_vl_paired_parser.add_argument('-fmut', '--forbidden_mut', help='List of string residues to ban for mutation, i.e. C M', nargs='*',
                              type=str, default=['C','M'])

    hum_vh_vl_paired_parser.add_argument('-Pairingdecrease', '--perc_allowed_decrease_pairing', help='Maximun ΔPairing score decrease allowed for a mutation',
                              type=float, default=1e-1)

    hum_vh_vl_paired_parser.add_argument('-a', '--a', help='Used for enhanced sampling method in multi-objective selection function: aΔVH+bΔPairing',
                              type=float, default=10)

    hum_vh_vl_paired_parser.add_argument('-b', '--b', help='Used for enhanced sampling method in multi-objective selection function: aΔVH+bΔPairing',
                              type=float, default=1)

    hum_vh_vl_paired_parser.add_argument('-pdb', '--pdb_file', help='Filepath to a pdb crystal structure of the nanobody of interest used to compute \
                                the solvent exposure. If the PDB is not very cleaned that might lead to some false results (which should be flagged by the \
                                program). If None, will predict the paired structure using ABodyBuilder2', type=str,
                              default=None)

    hum_vh_vl_paired_parser.add_argument('-ch_vh', '--ch_id_vh', help='PDB chain id of the heavy chain of interest. If -pdb is None, it does not matter', type=str,
                              default='H')

    hum_vh_vl_paired_parser.add_argument('-ch_vl', '--ch_id_vl', help='PDB chain id of the light chain of interest. If -pdb is None, it does not matter', type=str,
                              default='L')

    hum_vh_vl_paired_parser.set_defaults(func=lambda args: vh_vl_paired_humanisation.run(args))


    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



# (c) 2023 Sormannilab and Aubin Ramon
# 
# Vector-Quantization of the latent space in the AbNatiV model.
#
# Modified from (https://github.com/lucidrains/vector-quantize-pytorch/blob/master/vector_quantize_pytorch/vector_quantize_pytorch.py)
# under the copyright (c) 2020 Phil Wang (MIT Licenced)
# ============================================================================


from einops import rearrange
import pandas as pd
import torch
from torch import nn, einsum
import torch.nn.functional as F
import torch.distributed as distributed
from torch.amp import autocast

from .utils import uniform_init, kmeans, sample_vectors_distributed, noop, batched_embedding, batched_sample_vectors
from .utils import l2norm, gumbel_sample, ema_inplace, default
                           

class CosineSimCodebook(nn.Module):
    def __init__(
        self,
        dim,
        codebook_size,
        num_codebooks = 1,
        kmeans_init = False,
        kmeans_iters = 10,
        sync_kmeans = True,
        decay = 0.8,
        eps = 1e-5,
        threshold_ema_dead_code = 3,
        use_ddp = False,
        learnable_codebook = False,
        sample_codebook_temp = 0.
    ):
        super().__init__()
        self.decay = decay

        if not kmeans_init:
            embed = l2norm(uniform_init(num_codebooks, codebook_size, dim))
        else:
            embed = torch.zeros(num_codebooks, codebook_size, dim)

        self.codebook_size = codebook_size
        self.num_codebooks = num_codebooks

        self.kmeans_iters = kmeans_iters
        self.eps = eps
        self.threshold_ema_dead_code = threshold_ema_dead_code
        self.sample_codebook_temp = sample_codebook_temp

        self.sample_fn = sample_vectors_distributed if use_ddp and sync_kmeans else batched_sample_vectors
        self.kmeans_all_reduce_fn = distributed.all_reduce if use_ddp and sync_kmeans else noop
        self.all_reduce_fn = distributed.all_reduce if use_ddp else noop

        self.register_buffer('initted', torch.Tensor([not kmeans_init]))
        self.register_buffer('cluster_size', torch.zeros(num_codebooks, codebook_size))

        self.learnable_codebook = learnable_codebook
        if learnable_codebook:
            self.embed = nn.Parameter(embed)
        else:
            self.register_buffer('embed', embed)

    @torch.jit.ignore
    def init_embed_(self, data):
        if self.initted:
            return

        embed, cluster_size = kmeans(
            data,
            self.codebook_size,
            self.kmeans_iters,
            use_cosine_sim = True,
            sample_fn = self.sample_fn,
            all_reduce_fn = self.kmeans_all_reduce_fn
        )

        self.embed.data.copy_(embed)
        self.cluster_size.data.copy_(cluster_size)
        self.initted.data.copy_(torch.Tensor([True]))

    def replace(self, batch_samples, batch_mask):
        batch_samples = l2norm(batch_samples)

        for ind, (samples, mask) in enumerate(zip(batch_samples.unbind(dim = 0), batch_mask.unbind(dim = 0))):
            if not torch.any(mask):
                continue

            sampled = self.sample_fn(rearrange(samples, '... -> 1 ...'), mask.sum().item())
            self.embed.data[ind][mask] = rearrange(sampled, '1 ... -> ...')

    def expire_codes_(self, batch_samples):
        if self.threshold_ema_dead_code == 0:
            return

        expired_codes = self.cluster_size < self.threshold_ema_dead_code

        if not torch.any(expired_codes):
            return

        batch_samples = rearrange(batch_samples, 'h ... d -> h (...) d')
        self.replace(batch_samples, batch_mask = expired_codes)

    @autocast("cuda", enabled = False)
    def forward(self, x):
        needs_codebook_dim = x.ndim < 4

        x = x.float()

        if needs_codebook_dim:
            x = rearrange(x, '... -> 1 ...')

        shape, dtype = x.shape, x.dtype

        flatten = rearrange(x, 'h ... d -> h (...) d')
        flatten = l2norm(flatten)

        self.init_embed_(flatten)

        embed = self.embed if not self.learnable_codebook else self.embed.detach()
        embed = l2norm(embed)

        dist = einsum('h n d, h c d -> h n c', flatten, embed)
        embed_ind = gumbel_sample(dist, dim = -1, temperature = self.sample_codebook_temp)
        embed_onehot = F.one_hot(embed_ind, self.codebook_size).type(dtype)
        embed_ind = embed_ind.view(*shape[:-1])

        quantize = batched_embedding(embed_ind, self.embed)

        if self.training:
            bins = embed_onehot.sum(dim = 1)
            self.all_reduce_fn(bins)

            ema_inplace(self.cluster_size, bins, self.decay)

            zero_mask = (bins == 0)
            bins = bins.masked_fill(zero_mask, 1.)

            embed_sum = einsum('h n d, h n c -> h c d', flatten, embed_onehot)
            self.all_reduce_fn(embed_sum)

            embed_normalized = embed_sum / rearrange(bins, '... -> ... 1')
            embed_normalized = l2norm(embed_normalized)

            embed_normalized = torch.where(
                rearrange(zero_mask, '... -> ... 1'),
                embed,
                embed_normalized
            )

            ema_inplace(self.embed, embed_normalized, self.decay)
            self.expire_codes_(x)

        if needs_codebook_dim:
            quantize, embed_ind = map(lambda t: rearrange(t, '1 ... -> ...'), (quantize, embed_ind))

        return quantize, embed_ind


class VectorQuantize(nn.Module):
    def __init__(
        self,
        dim,
        codebook_size,
        codebook_dim,
        heads = 1,
        separate_codebook_per_head = False,
        decay = 0.8,
        eps = 1e-5,
        kmeans_init = True,
        kmeans_iters = 10,
        sync_kmeans = True,
        threshold_ema_dead_code = 3,
        commitment_weight = 1.,
        orthogonal_reg_weight = 0.,
        orthogonal_reg_active_codes_only = False,
        orthogonal_reg_max_codes = None,
        sample_codebook_temp = 0.,
        sync_codebook = False
    ):
        super().__init__()
        self.heads = heads
        self.separate_codebook_per_head = separate_codebook_per_head

        codebook_dim = default(codebook_dim, dim)
        codebook_input_dim = codebook_dim * heads

        requires_projection = codebook_input_dim != dim
        self.project_in = nn.Linear(dim, codebook_input_dim) if requires_projection else nn.Identity()
        self.project_out = nn.Linear(codebook_input_dim, dim) if requires_projection else nn.Identity()

        self.eps = eps
        self.commitment_weight = commitment_weight

        has_codebook_orthogonal_loss = orthogonal_reg_weight > 0
        self.orthogonal_reg_weight = orthogonal_reg_weight
        self.orthogonal_reg_active_codes_only = orthogonal_reg_active_codes_only
        self.orthogonal_reg_max_codes = orthogonal_reg_max_codes

        codebook_class = CosineSimCodebook

        self._codebook = codebook_class(
            dim = codebook_dim,
            num_codebooks = heads if separate_codebook_per_head else 1,
            codebook_size = codebook_size,
            kmeans_init = kmeans_init,
            kmeans_iters = kmeans_iters,
            sync_kmeans = sync_kmeans,
            decay = decay,
            eps = eps,
            threshold_ema_dead_code = threshold_ema_dead_code,
            use_ddp = sync_codebook,
            learnable_codebook = has_codebook_orthogonal_loss,
            sample_codebook_temp = sample_codebook_temp
        )

        self.codebook_size = codebook_size

    @property
    def codebook(self):
        codebook = self._codebook.embed
        if self.separate_codebook_per_head:
            return codebook

        return rearrange(codebook, '1 ... -> ...')

    def forward(self, x,):
        shape, device, heads, is_multiheaded, codebook_size = x.shape, x.device, self.heads, self.heads > 1, self.codebook_size

        x = self.project_in(x)

        if is_multiheaded:
            ein_rhs_eq = 'h b n d' if self.separate_codebook_per_head else '1 (b h) n d'
            x = rearrange(x, f'b n (h d) -> {ein_rhs_eq}', h = heads)

        quantize, embed_ind = self._codebook(x)

        if self.training:
            quantize = x + (quantize - x).detach()

        
        detached_inputs = x.detach()
        loss = F.mse_loss(quantize, detached_inputs, reduction='none')
        loss_pbe = torch.mean(loss, dim=(1,2)) # (batch_size)

        if self.commitment_weight > 0:
            detached_quantize = quantize.detach()
            commit_loss = F.mse_loss(detached_quantize, x, reduction='none')

            loss_pbe = loss_pbe + torch.mean(commit_loss * self.commitment_weight, dim=(1,2)) # (batch_size)

        if is_multiheaded:
            if self.separate_codebook_per_head:
                quantize = rearrange(quantize, 'h b n d -> b n (h d)', h = heads)
                embed_ind = rearrange(embed_ind, 'h b n -> b n h', h = heads)
            else:
                quantize = rearrange(quantize, '1 (b h) n d -> b n (h d)', h = heads)
                embed_ind = rearrange(embed_ind, '1 (b h) n -> b n h', h = heads)

        quantize_latent = quantize.detach().clone()
        quantize = self.project_out(quantize)

        avg_probs = torch.mean(F.one_hot(embed_ind, self.codebook_size).type(torch.float32).view((-1, self.codebook_size)), 0)
        perplexity = torch.exp(-torch.sum(avg_probs * torch.log(avg_probs + 1e-10)))

        return {
        'quantize_projected_in': x, # (batch_size, l_r, codebook_dim)
        'quantize_latent': quantize_latent, # (batch_size, l_r, codebook_dim)
        'quantize_projected_out': quantize, # (batch_size, l_r, dim)
        'loss_vq_commit_pbe': loss_pbe, # (batch_size)
        'perplexity': perplexity, # (batch_size)
        'encoding_indices': embed_ind # (batch_size, l_r)
    } 

